"""SVG icons."""

# Source: https://refokus-copy-to-clipboard.webflow.io/
CLONE = """
<svg class="icon icon-clone" width="24" height="24" viewBox="0 0 24 24" fill="none" stroke="#585B6C" stroke-width="2" xmlns="http://www.w3.org/2000/svg">
<path d="M17 6V5C17 3.34315 15.6569 2 14 2H6C4.34315 2 3 3.34315 3 5V15C3 16.6569 4.34315 18 6 18H7"/>
<rect x="7" y="6" width="14" height="16" rx="3"/>
</svg>
"""
CHECK = """
<svg class="icon icon-check" width="16" height="16" viewBox="0 0 16 16" fill="none" stroke="#28A745" strokeWidth="2" strokeLinecap="round" strokeLinejoin="round" xmlns="http://www.w3.org/2000/svg">
<path d="M13.25 4.75L6 12L2.75 8.75"/>
</svg>
"""
